/// This module contains a minimal definition of a MergeTag DataSet.
module mde.mergetag.dataset;

// package imports
import mde.mergetag.parse;
import mde.mergetag.exception;

// tango imports
import Util = tango.text.Util;
import tango.util.log.Log : Log, Logger;

/** Typedef for data & section indexes (can be changed to ulong if necessary.) */
typedef uint ID;

package struct MTFormatVersion {
    enum VERS : ubyte {	// convenient list of all known file format versions
        INVALID	= 0x00,
        MT01	= 0x01,		// not yet final
    }
    /// The current MergeTag version
    static const VERS Current = VERS.MT01;
    static const char[2] CurrentString = "01";
    
    static VERS parseString (char[] str)
    in {
            assert (str.length == 2);
    } body {
        if (str[0] == '0' && str[1] == '1') return VERS.MT01;
        else return VERS.INVALID;
    }
}

private Logger logger;
static this () {
    logger = Log.getLogger ("mde.mergetag.dataset");
}

/**************************************************************************************************
 * Data class; contains a DataSection class instance for each loaded section of a file.
 *
 * Any class implementing DataSection may be used to store data; by default a DefaultData class is
 * used when reading a file. Another class may be used by creating the sections before reading the
 * file or passing the reader a function to create the sections (see Reader.dataSecCreator).
 *
 * Could be a struct, except that structs are value types (not reference types).
 */
class DataSet
{
    DataSection header;			/// Header
    DataSection[ID] sec;		/// Dynamic array of sections
    
    /// Return a reference to the indexed item
    DataSection opIndex(ID i) {
        return sec[i];
    }
    
    /// Template to return all sections of a child-class type.
    T[ID] getSections (T : DataSection) () {
        T[ID] ret;
        foreach (ID id, DataSection s; sec) {
            T x = cast(T) s;
            if (x) ret[id] = x;	// if non-null
        }
        return ret;
    }
}

/**
 * Interface for data storage classes, which contain all data-tags loaded from a single section of a
 * file.
 *
 * A class implementing this may implement the addTag function to do whatever it likes with the
 * data passed; DefaultData separates this data out into supported types and stores it
 * appropriately, while throwing an error when unsupported types are passed, but a different
 * implementation could filter out the tags desired and use them directly, while ignoring the rest.
 * The parse module provides a useful set of templated functions to
 * convert the data accordingly. It is advised to keep the type definitions as defined in the file-
 * format except for user-defined types.
 *
 * Another idea for a DataSection class:
 * Use a void*[ID] variable to store all data (may also need a type var for each item).
 * addTag should call a templated function which calls parse then casts to a void* and stores the data.
 * Use a templated get(T)(ID) method which checks the type and casts to T. Small issue with this: storing
 * data in the file with an incorrect type could cause a lot of errors to be thrown in other code.
 */
interface DataSection
{
    /** Handles parsing of data items.
     *
     * Should throw an MTUnknownTypeException for unsupported types, after logging to logger.
     */
    void addTag (char[],ID,char[]);
    //void writeAll (Print!(char));	/// TBD
}

/**
 * Default DataSection class.
 *
 * Supports all the basic types currently supported and array versions of
 * each (except no arrays of binary or string types; these are already arrays).
 */
class DefaultData : DataSection
{
    //BEGIN DATA
    /** Data members for direct access. */
    bool	[ID]	Bool;
    byte	[ID]	Byte;	/// ditto
    short	[ID]	Short;	/// ditto
    int		[ID]	Int;	/// ditto
    long	[ID]	Long;	/// ditto
    ubyte	[ID]	UByte;	/// ditto
    ushort	[ID]	UShort;	/// ditto
    uint	[ID]	UInt;	/// ditto
    ulong	[ID]	ULong;	/// ditto
    
    bool[]	[ID]	BoolA;	/// ditto
    byte[]	[ID]	ByteA;	/// ditto
    short[]	[ID]	ShortA;	/// ditto
    int[]	[ID]	IntA;	/// ditto
    long[]	[ID]	LongA;	/// ditto
    ubyte[]	[ID]	UByteA;	/// ditto
    ushort[]	[ID]	UShortA;/// ditto
    uint[]	[ID]	UIntA;	/// ditto
    ulong[]	[ID]	ULongA;	/// ditto
    
    float	[ID]	Float;	/// ditto
    double	[ID]	Double;	/// ditto
    real	[ID]	Real;	/// ditto
    float[]	[ID]	FloatA;	/// ditto
    double[]	[ID]	DoubleA;/// ditto
    real[]	[ID]	RealA;	/// ditto
    
    char	[ID]	Char;	/// ditto
    char[]	[ID]	CharA;	/// ditto
    
    /** Alias names */
    alias	CharA	String;
    alias	UByteA	Binary;	/// ditto
    //END DATA
    
    /+ Could use this:
    private template addTagTp(alias Var, T) (ID id, char[] dt) {
        Var[id] = parse!(T) (dt);
    }
    or a mixin.
    +/
    // Unfortunately, I think each case needs to be mentioned explicitly to tie it to the correct
    // data member.
    void addTag (char[] tp, ID id, char[] dt) {	/// for adding tags
        try {
            postTrim (tp);
            // parse tp, then use if statements to replace the following switch
            
        switch(Util.trim(tp)) {
            case "1":
            case "bool":
                Arg!(bool)[id] = parse!(bool) (dt);
            break;
            case "s8":
            case "byte":
                addTagTp!(byte) (id, dt);
            break;
            case "s16":
            case "short":
                Short[id] = parse!(short) (dt);
            break;
            case "s32":
            case "int":
                Arg!(int)[id] = parse!(int) (dt);
            break;
            case "s64":
            case "long":
                Long[id] = parse!(long) (dt);
            break;
            case "u8":
            case "ubyte":
                UByte[id] = parse!(ubyte) (dt);
            break;
            case "u16":
            case "ushort":
                UShort[id] = parse!(ushort) (dt);
            break;
            case "u32":
            case "uint":
                UInt[id] = parse!(uint) (dt);
            break;
            case "u64":
            case "ulong":
                ULong[id] = parse!(ulong) (dt);
            break;
            
            case "1[]":
            case "bool[]":
                BoolA[id] = parse!(bool[]) (dt);
            break;
            case "s8[]":
            case "byte[]":
                ByteA[id] = parse!(byte[]) (dt);
            break;
            case "s16[]":
            case "short[]":
                ShortA[id] = parse!(short[]) (dt);
            break;
            case "s32[]":
            case "int[]":
                IntA[id] = parse!(int[]) (dt);
            break;
            case "s64[]":
            case "long[]":
                LongA[id] = parse!(long[]) (dt);
            break;
            case "u8[]":
            case "ubyte[]":
            case "binary":
                UByteA[id] = parse!(ubyte[]) (dt);
            break;
            case "u16[]":
            case "ushort[]":
                UShortA[id] = parse!(ushort[]) (dt);
            break;
            case "u32[]":
            case "uint[]":
                UIntA[id] = parse!(uint[]) (dt);
            break;
            case "u64[]":
            case "ulong[]":
                ULongA[id] = parse!(ulong[]) (dt);
            break;
            
            case "fp32":
            case "float":
                Float[id] = parse!(float) (dt);
            break;
            case "fp64":
            case "double":
                Double[id] = parse!(double) (dt);
            break;
            case "fp":
            case "real":
                Real[id] = parse!(real) (dt);
            break;
            case "fp32[]":
            case "float[]":
                FloatA[id] = parse!(float[]) (dt);
            break;
            case "fp64[]":
            case "double[]":
                DoubleA[id] = parse!(double[]) (dt);
            break;
            case "fp[]":
            case "real[]":
                RealA[id] = parse!(real[]) (dt);
            break;
            
            case "UTF8":
            case "char":
                Char[id] = parse!(char) (dt);
            break;
            case "UTF8[]":
            case "char[]":
            case "string":
                CharA[id] = parse!(char[]) (dt);
            break;
            
            default:
                logger.warn ("Type not supported: " ~ tp);
                throw new MTUnknownTypeException ();
        }
        } catch (Exception e) {
            // an error should have already been logged
        }
    }
    
private:
    void addTagTp(T) (ID id, char[] dt) {
        Arg!(T)[id] = parse!(T) (dt);
    }
    
    // use as: mixin Arg!(type); or Arg!(type)
    template Arg(T : bool) {
        alias Bool Arg;
    }
    template Arg(T : byte) {
        alias Byte Arg;
    }
    template Arg(T : short) {
        alias Short Arg;
    }
    template Arg(T : int) {
        alias Int Arg;
    }
    template Arg(T : long) {
        alias Long Arg;
    }
    template Arg(T : ubyte) {
        alias UByte Arg;
    }
    template Arg(T : ushort) {
        alias UShort Arg;
    }
    template Arg(T : uint) {
        alias UInt Arg;
    }
    template Arg(T : ulong) {
        alias ULong Arg;
    }
    template Arg(T : bool[]) {
        alias BoolA Arg;
    }
    template Arg(T : byte[]) {
        alias ByteA Arg;
    }
    template Arg(T : short[]) {
        alias ShortA Arg;
    }
    template Arg(T : int[]) {
        alias IntA Arg;
    }
    template Arg(T : long[]) {
        alias LongA Arg;
    }
    template Arg(T : ubyte[]) {
        alias UByteA Arg;
    }
    template Arg(T : ushort[]) {
        alias UShortA Arg;
    }
    template Arg(T : uint[]) {
        alias UIntA Arg;
    }
    template Arg(T : ulong[]) {
        alias ULongA Arg;
    }
    template Arg(T : float) {
        alias Float Arg;
    }
    template Arg(T : double) {
        alias Double Arg;
    }
    template Arg(T : real) {
        alias Real Arg;
    }
    template Arg(T : float[]) {
        alias FloatA Arg;
    }
    template Arg(T : double[]) {
        alias DoubleA Arg;
    }
    template Arg(T : real[]) {
        alias RealA Arg;
    }
    template Arg(T : char) {
        alias Char Arg;
    }
    template Arg(T : char[]) {
        alias CharA Arg;
    }
}

/+
class TemplateData : DataSection
{
    void addTag (char[] tp, ID id, char[] dt) {
        // runtime deduction of tp and aliasing?
        // CANNOT add data at runtime though.
    }
    // will this work? no idea.
    // templates can't be used to add non-static elements, so use a static array at index: this
    template Data(T,TemplateData* p) {
        static T[ID][TemplateData*] Data;
    }
}
+/